--
-- StdDIS for Hugs
--
-- (c) Thomas Nordin and Alastair Reid, 1997
--

module Xlib_StdDIS
        ( StablePtr
        , module Int
        , module Addr
        , module IOExts
        , MbString
        , marshall_bool_,      unmarshall_bool_
        , marshall_string_,    unmarshall_string_
        , marshall_stringLen_, unmarshall_stringLen_
        , makeStablePtr, deRefStablePtr, freeStablePtr
        , malloc, free
	, word8ToWord32, word32ToWord8, word16ToWord32, word32ToWord16 
	, int8ToInt, intToInt8, int16ToInt, intToInt16, int32ToInt, intToInt32
	, ForeignObj
	, makeForeignObj
	, malloc, free
        ) where


import Int
import Hugs.Word
import Addr
import IOExts
import Monad( zipWithM_ )
import Hugs.StablePtr ( StablePtr, 
                 makeStablePtr, deRefStablePtr, 
                 freeStablePtr, 
               )
import Foreign hiding (malloc, free)
import Ptr

word8ToWord32 :: Word8 -> Word32
word8ToWord32 x = toEnum (fromEnum x)

word32ToWord8 :: Word32 -> Word8
word32ToWord8 x = toEnum (fromEnum x)

word16ToWord32 :: Word16 -> Word32
word16ToWord32 x = toEnum (fromEnum x)

word32ToWord16 :: Word32 -> Word16
word32ToWord16 x = toEnum (fromEnum x)


int8ToInt :: Int8 -> Int
int8ToInt x = fromEnum x

intToInt8 :: Int -> Int8
intToInt8 x = toEnum x

int16ToInt :: Int16 -> Int
int16ToInt x = fromEnum x

intToInt16 :: Int -> Int16
intToInt16 x = toEnum x

int32ToInt :: Int32 -> Int
int32ToInt x = fromEnum x

intToInt32 :: Int -> Int32
intToInt32 x = toEnum x






marshall_bool_ :: Bool -> IO Int
marshall_bool_ True  = return 1
marshall_bool_ False = return 0

unmarshall_bool_ :: Int -> IO Bool
unmarshall_bool_ 0 = return False
unmarshall_bool_ _ = return True

-- Ignore "IO" part of result type

----------------------------------------------------------------
-- Strings
----------------------------------------------------------------


type MbString      = Maybe String

marshall_string_ :: [Char] -> IO Addr
marshall_string_ cs =
 do arr <- allocCharStar (1 + length cs)
    zipWithM_ (writeCharAddr arr) [0..] (cs ++ "\0")
    return arr

marshall_stringLen_ :: [Char] -> IO (Addr, Int)
marshall_stringLen_ cs =
 do let l = length cs
    arr <- allocCharStar (l+1)
    zipWithM_ (writeCharAddr arr) [0..] (cs ++ "\0")
    return (arr,l)

unmarshall_string_ :: Addr -> IO String
unmarshall_string_ ptr = reads "" 0
  where
    reads str i =
      readCharAddr ptr i >>= \c ->
      if c == '\0' then
        return (reverse str)
      else
        reads (c:str) (i+1)

unmarshall_stringLen_ :: Addr -> Int -> IO String
unmarshall_stringLen_ ptr l = mapM (readCharAddr ptr) [0..l-1]

writeCharAddr :: Addr -> Int -> Char -> IO ()
writeCharAddr s i v =
  prim_writeCharAddr s i v
foreign import  ccall unsafe "Xlib_StdDIS_stub_ffi.h prim_writeCharAddr" prim_writeCharAddr :: Addr -> Int -> Char -> IO ()
 
readCharAddr :: Addr -> Int -> IO Char
readCharAddr s i =
  prim_readCharAddr s i
  >>= \  res1  ->
  (return (res1))
foreign import  ccall unsafe "Xlib_StdDIS_stub_ffi.h prim_readCharAddr" prim_readCharAddr :: Addr -> Int -> IO (Char)

allocCharStar :: Int -> IO Addr
allocCharStar arg1 =
  prim_allocCharStar arg1
  >>= \  res1  ->
  (return (res1))
foreign import  ccall unsafe "Xlib_StdDIS_stub_ffi.h prim_allocCharStar" prim_allocCharStar :: Int -> IO (Addr)

----------------------------------------------------------------
-- malloc/free
----------------------------------------------------------------


malloc :: Word32 -> IO Addr
malloc arg1 =
  prim_malloc arg1
  >>= \ gc_result ->
  access_prim_malloc_res1 gc_result >>= \ res1 ->
  access_prim_malloc_gc_failed gc_result >>= \ gc_failed ->
  access_prim_malloc_gc_failstring gc_result >>= \ gc_failstring ->
  if ( gc_failed /= (0::Int))
  then unmarshall_string_ gc_failstring >>=  ioError  . userError
  else (return (res1))
foreign import  ccall unsafe "Xlib_StdDIS_stub_ffi.h prim_malloc" prim_malloc :: Word32 -> IO (Ptr r1)
foreign import ccall unsafe "Xlib_StdDIS_stub_ffi.h" access_prim_malloc_res1 :: Ptr a1 -> IO (Addr)
foreign import ccall unsafe "Xlib_StdDIS_stub_ffi.h" access_prim_malloc_gc_failed :: Ptr a1 -> IO (Int)
foreign import ccall unsafe "Xlib_StdDIS_stub_ffi.h" access_prim_malloc_gc_failstring :: Ptr a1 -> IO (Addr)
 
free :: Addr -> IO ()
free arg1 =
  prim_free arg1
foreign import  ccall unsafe "Xlib_StdDIS_stub_ffi.h prim_free" prim_free :: Addr -> IO ()

 
----------------------------------------------------------------
-- Stable pointers
----------------------------------------------------------------

--
-- Use "stable" to create a stable pointer
-- 
-- Use "stablePtr" to manipulate (previously constructed) stable pointers 
-- in Haskell.
--

----------------------------------------------------------------
-- Foreign Objects
----------------------------------------------------------------

-- Use foreign when you want to control what free routine
-- to use.
--

type ForeignObj = ForeignPtr ()

makeForeignObj :: Ptr () -> Ptr () -> IO ForeignObj
makeForeignObj a f = newForeignPtr (castPtrToFunPtr f) a

----------------------------------------------------------------
-- End of StdDIS
----------------------------------------------------------------

